#!/usr/bin/env python

import argparse
import os
import sys
from threading import Thread
import subprocess
import datetime
import shutil

try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty  # python 3.x

ON_POSIX = 'posix' in sys.builtin_module_names


def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line.rstrip())
    out.close()
    queue.put('<END>')


def main():

    script_home = os.path.dirname(os.path.realpath(__file__))
    models = os.path.realpath(script_home + '/../models')
    qsspn_bin_dir = os.path.realpath(script_home + '/../v3.3/src/bin/')

    parser = argparse.ArgumentParser(description="Utility for running QSSPN.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-m', '--mode', default='standalone',
                        choices=['standalone', 'server', 'client', 'proxy', 'thrift_client'],
                        help='QSSPN running mode.')
    parser.add_argument('-c', '--cache-size', default=1000, type=int,  # ~ 1GB
                        help='Maximum cache size (in MB).')
    parser.add_argument('--use-memcheck', default=False,
                        action='store_true',
                        help='Run under memcheck to find memory leaks '
                        '(requires valgrind).')

    parser.add_argument('model', help='Name of model to run', nargs='?',
                        default='BileAcidHomeostasis1.0')

    parser.add_argument('-s', '--server-address', default='127.0.0.1',
                        help='QSSPN server address for client-server mode.')
    parser.add_argument('-p', '--server-port', default='30130',
                        help='QSSPN server port for client-server mode.')
    parser.add_argument('-t', '--thrift-port', default='9090',
                        help='QSSPN thrift port.')

    parser.add_argument('--model-dir', default=models,
                        help='Path to the directory with model files (.ctrl)')
    parser.add_argument('--qsspn-binary-dir', default=qsspn_bin_dir,
                        help='Path to the directory containing QSSPN binaries')
    parser.add_argument('-w', '--workspace', default='workspace_{mode}_{datetime}',
                        help='The name of the folder where qsspn will be run.')
    parser.add_argument('--ld-path', help='Extra dynamic library path')

    parser.add_argument('--initial-cache', help='File to read cache from')
    parser.add_argument('--save-cache-to', help="File to save cache to when it's almost full")

    args = parser.parse_args()

    model_dir = os.path.realpath(args.model_dir)
    bin_dir = os.path.realpath(args.qsspn_binary_dir)

    qsspn_file = model_dir + '/' + args.model + '.qsspn'
    ctrl_file = model_dir + '/' + args.model + '.ctrl.txt'
    binary = bin_dir + '/qsspn_' + args.mode
    if args.mode == 'thrift_client':
        binary = bin_dir + '/qsspn_example_thrift_client'
    command = [binary, qsspn_file, ctrl_file]
    if args.mode in ['server', 'client', 'proxy']:
        command += [args.server_address, args.server_port]
    if args.mode == 'server':
        command += [str(args.cache_size * 1024 * 1024 / 4)]
    if args.mode == 'proxy':
        command += [str(args.thrift_port)]

    if args.use_memcheck:
        command = ['valgrind', '--log-file=qsspn.valgrind.%p.log'] + command



    if args.save_cache_to:
        initial_cache = args.initial_cache or ""
        cache_args = map(os.path.realpath, [initial_cache, args.save_cache_to])
    elif args.initial_cache:
        cache_args = [os.path.realpath(args.initial_cache)]
    else:
        cache_args = []

    if args.mode == 'server':
        command += cache_args
    print('Running subprocess: `' + ' '.join(command) + '` ...')

    workspace = args.workspace
    workspace_params = {}
    if '{datetime}' in workspace:
        datestr = '{:%Y-%m-%d_%H:%M:%S}'.format(datetime.datetime.now())
        workspace_params['datetime'] = datestr

    if '{mode}' in workspace:
        workspace_params['mode'] = args.mode
    if len(workspace_params):
        workspace = workspace.format(**workspace_params)
    if not os.path.exists(workspace):
        os.mkdir(workspace)
    outpath = workspace + '/output.xls'
    if os.path.exists(outpath):
        # qsspn appends results, this is bad
        print('Warning: output file {out} already exists - removing...'.format(out=outpath))
        os.remove(outpath)


    start_time = datetime.datetime.now()

    env = os.environ.copy()
    if args.ld_path:
       env['LD_LIBRARY_PATH'] = os.path.realpath(args.ld_path)
    qsspn = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             cwd=workspace, bufsize=1, close_fds=True, env=env)

    q = Queue()
    t = Thread(target=enqueue_output, args=(qsspn.stdout, q))
    terr = Thread(target=enqueue_output, args=(qsspn.stderr, q))
    terr.daemon = True
    t.daemon = True
    t.start()
    terr.start()

    print(' === QSSPN running in {ws} (mode: {mode}): ==='.format(ws=workspace, mode=args.mode))
    while True:
        line = q.get()
        if line == '<END>':
            break
        print(line)
    end_time = datetime.datetime.now()
    print(' === QSSPN {} run finished ==='.format(args.mode))
    print('Wall clock run time: {}'.format(end_time - start_time))
    if args.mode == 'server':
        print('Note that QSSPN was run in server mode, so wall clock time '
              'is probably meaningless here - check client running time instead')
    print('Results saved to {}'.format(workspace))


if __name__ == "__main__":
    main()
