/*
 *  PetriNetAlgorithms.cpp
 *  pnfba_xcode
 *
 *  Created by Andrzej Kierzek on 06/01/2011.
 *  Copyright 2011 Andrzej Kierzek. All rights reserved.
 *
 */

#include "PetriNet.h"
#include "ModeConfiguration.h"
#include "metabolism/EnzymesTokensEncoder.h"
#include "metabolism/BinaryEnzymesTokensEncoder.h"
#include "metabolism/SubstanceInterface.h"
#include <stdexcept>

double Interaction::activity(int i) {
    double result = 0;
    if (activities[i].isMassAction()) {
        result = 1;
        double fact = 1;
        for (int j = 0; j < (int) stoich[i]; j++) {
            result *= (substrates[i]->getState() - j);
            fact *= (j + 1);
        }
        return (result / fact);
    }
    for (int j = 0; j < activities[i].size; j++) {
        result = (substrates[i]->getState() >= activities[i].getThreshold(j)) ? activities[i].getActivity(j) : result;
    }
    return result;
}

double Interaction::propensity(DriverInterface *gsmn) {
    a = c;
    for (int i = 0; i < Ns; i++) a *= activity(i);
    for (int i = 0; i < Nm; i++) metabolites[i]->metaboliteAction(gsmn);
    if (a != 0 && a != 1 ) {
      cout << "Non-bool propensity (with driver) in " << name << ": " << a;
    }
    return a;
}

double Interaction::propensity() {
    if (propensity_expression) {
        for (int i = 0; i < Ns; i++) activity_values[i] = activity(i);
        a = propensityFunction.evaluate();
        if (a != 0 && a != 1) {
          cout << "Non-bool propensity (with expression) in " << name << ": " << a << endl;
        }
        return (a);
    }
    a = c;
    for (int i = 0; i < Ns; i++) a *= activity(i);
    if (a != 0 && a != 1) {
      cout << "Non-bool propensity (without expression) in " << name << ": " << a << endl;
    }
    return a;

}

void Interaction::stateChange() {
    for (int i = 0; i < Nc; i++) consumed[i]->stateChange += (-1) * a;
    for (int i = 0; i < Np; i++) products[i]->stateChange += a;
}

void Interaction::elementaryChange() {
    for (int i = 0; i < Nc; i++) consumed[i]->stateChange = (-1) * stoich[i];
    for (int i = 0; i < Np; i++) products[i]->stateChange = pstoich[i];
}

void Interaction::execute() {
    stateChange();
    for (int i = 0; i < Nc; i++) consumed[i]->update();
    for (int i = 0; i < Np; i++) products[i]->update();
}

void Interaction::fire() {
    for (int i = 0; i < Nc; i++) consumed[i]->update(-1 * stoich[i]);
    for (int i = 0; i < Np; i++) products[i]->update(pstoich[i]);
}

void Interaction::set() {
    double value;
    value = 1.0;
    if (reset_expression) {
        for (int i = 0; i < Ns; i++) activity_values[i] = activity(i);
        value = resetFunction.evaluate();
    }
    for (int i = 0; i < Np; i++) {
        if (products[i]->resetDone) {
            cout << "Two reset transitions cannot act on place " << products[i]->getName() << " at the same time" << endl;
            exit(0);
        }
        products[i]->resetValue = value * pstoich[i];
        products[i]->resetDone = true;
    }
}

//void Interaction::flux(Driver *gsmn) {
//    double flux;
//    flux = gsmn->getLP()->getColValue(flux_request);
//    for (int i = 0; i < Np; i++) products[i]->setState(flux);
//}

void PetriNet::synchronousUpdate() {
    int i;

    for (i = 0; i < Ne; i++) enzymes[i]->enzymeAction(gsmnDriver);
    for (i = 0; i < Nm; i++) metabolites[i]->fbaRequired(true);
    for (i = 0; i < Ni; i++) {
        interactions[i].propensity(gsmnDriver);
        interactions[i].stateChange();
    }
    for (i = 0; i < Ns; i++) substances[i].update();
}

void PetriNet::prepareForSampling(int ns, int sd) {
    Nsample = ns;
    trajectories.clear();
    Nia = 0;
    if (rnd == NULL) {
      rnd = new RandomNumberGenerator(sd);
    } else {
      rnd->rmarzam(sd);
    }
}

string PetriNet::asynchronousUpdate() {
    Nia = 0;
    active.clear();
    fbaRequired = false;
    for (int k = 0; k < Ne; k++) if (enzymes[k]->enzymeAction(gsmnDriver)) fbaRequired = true;
    if (fbaRequired) for (int k = 0; k < Nm; k++) metabolites[k]->fbaRequired(true);
    for (int j = 0; j < Ni; j++) {
        interactions[j].propensity(gsmnDriver);
        if (interactions[j].a > 0) {
            active.push_back(&interactions[j]);
            Nia++;
        }
    }
    int next = rnd->uniform_index(Nia);
    active[next]->execute();
    return active[next]->getName();
}

/* Test function for nutrient dynamics. */
//void PetriNet::NutrientDynamics(double dt, int max) {
//    int j;
//
//    cout << objective->getName() << "\t";
//    for (j = 0; j < Nn; j++) cout << nutrients[j]->getName() << "\t";
//    cout << objective->getName() << "\t";
//    for (j = 0; j < Nn; j++) cout << nutrients[j]->getName() << "\t";
//    cout << endl;
//
//    buildGsmnDriver();
//
//    for (int i = 0; i < max; i++) {
//        for (j = 0; j < Ne; j++) enzymes[j]->initialEnzymeAction(gsmnDriver);
//        for (j = 0; j < Nn; j++) nutrients[j]->setNutrientBounds(gsmnDriver);
//        objective->objectiveAction(gsmnDriver);
//        for (j = 0; j < Nn; j++) nutrients[j]->nutrientState(gsmnDriver, dt);
//        for (j = 0; j < Nn; j++) nutrients[j]->setNutrientBounds(gsmnDriver);
//
//        cout << objective->getFlux() << "\t";
//        for (j = 0; j < Nn; j++) cout << nutrients[j]->getFlux() << "\t";
//        cout << objective->getState() << "\t";
//        for (j = 0; j < Nn; j++) cout << nutrients[j]->getState() << "\t";
//        cout << endl;
//    }
//}

/************************************************************************************/
/*                        MAIN SIMULATION ENGINE                                    */

/************************************************************************************/

string PetriNet::ssaUpdate() {
    double a0, r2, av;
    int next;

    a0 = 0;
    for (int j = 0; j < Ni; j++) a0 += interactions[j].propensity();
    if (a0 == 0) return (string(NOEVENT));
    deltaTime = 1 / a0;
    Time += deltaTime;
    r2 = rnd->marzam();
    av = 0;
    next = 0;
    while (true) {
        av += interactions[next].a;
        if (av >= a0 * r2) break;
        next++;
    }
    interactions[next].fire();
    return interactions[next].getName();

}

void PetriNet::splitReactions() {
    Ninstant = 0;
    Nfast = 0;
    Nslow = 0;
    int is, fs, ins, rs;
    for (int i = 0; i < Ni; i++)
        if (interactions[i].slow) {
            Nslow++;
        } else if (interactions[i].instant) {
            Ninstant++;
        } else if (interactions[i].reset) {
            Nreset++;
        } else if (interactions[i].fba_flux) {
        } else {
            Nfast++;
        }

    if (slow.size() != 0 || fast.size() != 0 || instant.size() != 0
        || reset.size() != 0) {
          cerr << "non-zero interaction vector sizes "
                  " - probably splitReactions called twice\n";
          slow.clear();
          fast.clear();
          instant.clear();
          reset.clear();
    }
    for (int i = 0; i < Ni; i++) if (interactions[i].slow) {
            slow.push_back(&interactions[i]);
        } else if (interactions[i].instant) {
            instant.push_back(&interactions[i]);
        } else if (interactions[i].reset) {
            reset.push_back(&interactions[i]);
        } else if (interactions[i].fba_flux) {
            interactions[i].setFluxProducts();
        } else {
            fast.push_back(&interactions[i]);
        }

}

void PetriNet::updateFastReactions1() {
    int i;
    string name;
    double buff;

    for (i = 0; i < Nslow; i++) slow[i]->a = 0;
    for (i = 0; i < Nfast; i++) {
        fast[i]->propensity();
        if (fast[i]->instant) {
            fast[i]->a = (fast[i]->a == 0) ? 0 : 1;
        } else {
            fast[i]->a *= deltaTime;
            if (msg_flag) {
                name = fast[i]->getName();
                buff = fast[i]->a;
                cout << "Fast reaction: " + name + " " << buff << endl;
            }
        }
        fast[i]->stateChange();
    }
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4)
            substances[i].update();
}

void PetriNet::updateInstantReactions() {
    int i, next;
    double r, av, a0;
    vector<int> enabled;
    vector<double> propensities;
    vector<int> queue;

    a0 = 0;
    for (i = 0; i < Ninstant; i++) {
        instant[i]->propensity();
        a0 += instant[i]->a;
        if (instant[i]->a > 0) {
            enabled.push_back(i);
            propensities.push_back(instant[i]->a);
        }
    }

    while (!enabled.empty()) {
        r = rnd->marzam();
        av = 0;
        next = 0;
        while (true) {
            av += instant[enabled[next]]->a;
            if (av >= a0 * r) break;
            next++;
        }
        queue.push_back(enabled[next]);
        enabled.erase(enabled.begin() + next);
        a0 -= propensities[next];
        propensities.erase(propensities.begin() + next);
    }

    for (i = 0; i < queue.size(); i++) {
        instant[queue[i]]->propensity();
        if (instant[queue[i]]->a > 0) instant[queue[i]]->fire();
    }

    enabled.clear();
    queue.clear();

}

void PetriNet::updateResetReactions() {
    int i;

    for (i = 0; i < Nreset; i++) {
        reset[i]->propensity();
        if (reset[i]->a > 0) reset[i]->set();
    }
    for (int i = 0; i < Ns; i++) {
        if (substances[i].resetDone)
            substances[i].setState(substances[i].resetValue);
        substances[i].resetDone = false;
    }
}

void PetriNet::updateFastReactions2() {
    /* Naive shortest timestep algorithm for adaptive timestep in Euler ODE integration */
    int i;
    string name;
    double dt, dt1;

    for (i = 0; i < Nslow; i++) slow[i]->a = 0;
    for (i = 0; i < Ninstant; i++) instant[i]->a = 0;
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4)
            substances[i].stateChange = 0;

    // Calculate propensities of fast and instant reactions and pass propenisties as state changes
    for (i = 0; i < Nfast; i++) {
        fast[i]->propensity();
        fast[i]->stateChange();
    }

    // Find minimal timestep such that maximal change is maxChange
    dt = TIME_INF;
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4) {
            if (substances[i].stateChange == 0 || substances[i].getState() == 0) {
                substances[i].stateChange = 0;
                continue;
            }
            dt1 = maxChange * substances[i].getState() / abs(substances[i].stateChange);
            substances[i].stateChange = 0;
            if (dt1 < dt && dt1 > MIN_DT) dt = dt1;
        }

    // set timestep
    deltaTime = (dt > maxTimestep) ? maxTimestep : dt;

    // re-calculate state changes of fast reactions
    for (i = 0; i < Nfast; i++) {
        fast[i]->a *= deltaTime;
        fast[i]->stateChange();
    }

    // update all substances (propensties of slow reactions are still 0)
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4)
            substances[i].update();
}

void PetriNet::updateFastReactions3(double maxdt) {
    /* Naive shortest timestep algorithm for adaptive timestep in Euler ODE integration */
    int i, maxs;
    string name;
    double dt, dt1, sc1, sc;

    for (i = 0; i < Nslow; i++) slow[i]->a = 0;
    for (i = 0; i < Ninstant; i++) instant[i]->a = 0;
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4)
            substances[i].stateChange = 0;

    // Calculate propensities of fast and instant reactions and pass propenisties as state changes
    for (i = 0; i < Nfast; i++) {
        fast[i]->propensity();
        fast[i]->stateChange();
    }

    // Find minimal timestep such that maximal change is maxChange
    dt = TIME_INF;
    maxs = 0;
    sc1 = 0;
    sc = 0;
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4) {
            if (substances[i].stateChange == 0 || substances[i].getState() == 0) {
                substances[i].stateChange = 0;
                continue;
            }
            sc1 = substances[i].stateChange;
            name = substances[i].getName();
            dt1 = maxChange * substances[i].getState() / abs(substances[i].stateChange);
            substances[i].stateChange = 0;
            if (dt1 < dt && dt1 > MIN_DT) {
                dt = dt1;
                sc = sc1;
                maxs = i;
            }
        }

    if (print_max_change)
        cout << "PRINT_MAX_CHANGE Time:" << " " << Time << " Substance:" << substances[maxs].getName() << " Rate:" << sc << endl;

    // set timestep
    deltaTime = (dt > maxdt) ? maxdt : dt;

    // re-calculate state changes of fast reactions
    for (i = 0; i < Nfast; i++) {
        fast[i]->a *= deltaTime;
        fast[i]->stateChange();
    }

    // update all substances (propensties of slow reactions are still 0)
    for (i = 0; i < Ns; i++)
        if (substances[i].getType() < 4)
            substances[i].update();
}

int PetriNet::fireDelayedReaction() {
    int next, eridx;
    double t;

    next = (-1);
    for (int j = 0; j < delays.size(); j++) {
        t = delays[j];
        if (t < Time) {
            next = delayed[j];
            Time = delays[j];
            eridx = j;
        }
    }
    if (next > (-1)) {
        slow[next]->fire();
        delays.erase(delays.begin() + eridx);
        delayed.erase(delayed.begin() + eridx);
        return next;
    }
    return (-1);
}

string PetriNet::hybridUpdate3() {
    double a0, r1, r2, av, tend, tstart, maxdt;
    int next;

    a0 = 0;
    for (int j = 0; j < Nslow; j++) a0 += slow[j]->propensity();
    if (a0 == 0) {
        updateFastReactions3(maxTimestep);
        Time += deltaTime;
        if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
        return (string(NOEVENT));
    }
    r1 = rnd->marzam();
    deltaTime = (1 / a0) * log(1 / r1);
    if (deltaTime > maxTimestep) {
        tend = Time + maxTimestep;
        while (Time < tend) {
            updateFastReactions3(maxTimestep);
            Time += deltaTime;
        }
        if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
        return (NOEVENT);
    }
    tstart = Time;
    Time += deltaTime;
    maxdt = deltaTime;
    if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
    r2 = rnd->marzam();
    av = 0;
    next = 0;
    while (true) {
        av += slow[next]->a;
        if (av >= a0 * r2) break;
        next++;
    }
    while (tstart < Time) {
        updateFastReactions3(maxdt);
        tstart += deltaTime;
    }
    if (slow[next]->delay > 0) {
        delayed.push_back(next);
        delays.push_back(Time + slow[next]->delay);
        deltaTime = maxTimestep / 1000;
        return ("delayed:" + slow[next]->getName());
    }
    slow[next]->fire();
    return slow[next]->getName();
}

string PetriNet::hybridUpdate2() {
    double a0, r1, r2, av, tend;
    int next;

    a0 = 0;
    for (int j = 0; j < Nslow; j++) a0 += slow[j]->propensity();
    if (a0 == 0) {
        updateFastReactions2();
        Time += deltaTime;
        if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
        return (string(NOEVENT));
    }
    r1 = rnd->marzam();
    deltaTime = (1 / a0) * log(1 / r1);
    if (deltaTime > maxTimestep) {
        tend = Time + maxTimestep;
        while (Time < tend) {
            updateFastReactions2();
            Time += deltaTime;
        }
        if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
        return (NOEVENT);
    }
    Time += deltaTime;
    if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
    r2 = rnd->marzam();
    av = 0;
    next = 0;
    while (true) {
        av += slow[next]->a;
        if (av >= a0 * r2) break;
        next++;
    }
    if (slow[next]->delay > 0) {
        delayed.push_back(next);
        delays.push_back(Time + slow[next]->delay);
        deltaTime = maxTimestep / 1000;
        return ("delayed:" + slow[next]->getName());
    }
    slow[next]->fire();
    return slow[next]->getName();
}

string PetriNet::hybridUpdate1() {
    double a0, r1, r2, av;
    int next;

    a0 = 0;
    for (int j = 0; j < Nslow; j++) a0 += slow[j]->propensity();
    if (a0 == 0) {
        Time += deltaTime;
        if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
        return (string(NOEVENT));
    }
    r1 = rnd->marzam();
    deltaTime = (1 / a0) * log(1 / r1);
    if (deltaTime > maxTimestep) {
        deltaTime = maxTimestep;
        Time += deltaTime;
        if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
        return (NOEVENT);
    }
    Time += deltaTime;
    if ((next = fireDelayedReaction())>(-1)) return (slow[next]->getName());
    r2 = rnd->marzam();
    av = 0;
    next = 0;
    while (true) {
        av += slow[next]->a;
        if (av >= a0 * r2) break;
        next++;
    }
    if (slow[next]->delay > 0) {
        delayed.push_back(next);
        delays.push_back(Time + slow[next]->delay);
        deltaTime = maxTimestep / 1000;
        return ("delayed:" + slow[next]->getName());
    }
    slow[next]->fire();
    return slow[next]->getName();
}

void PetriNet::loadMetabolismStep(const ModeConfiguration::Config& config) {
    if (encoder == NULL) {
    	encoder = new BinaryEnzymesTokensEncoder(enzymes);
    }
	int encoding = encoder->encode();
	vector<double>* metaboliteStates = metabolismCache->get(encoding);
	for (int k = 0; k < Nm; k++) {
		metabolites[k]->setState(metaboliteStates->at(k));
	}
}

void PetriNet::runMetabolismStep(const ModeConfiguration::Config& config) {
	for (int k = 0; k < Ne; k++) {
		enzymes[k]->setEnzymeBounds(gsmnDriver);
	}
	switch (config.mode) {
        case ModeConfiguration::STANDALONE:
            for (int k = 0; k < Nm; k++) {
                metabolites[k]->producibilityStateStandalone(gsmnDriver);
            }
            break;
        case ModeConfiguration::CLIENT:
            for (int k = 0; k < Nm; k++) {
                metabolites[k]->producibilityStateBuffClient(gsmnDriver);
            }
            break;
        default:
            throw std::runtime_error("Unknown running mode");
    }
}

/*****************************************************************************************/
/*                           WORKHORSE SIMULATION ROUTINE                                */

/*****************************************************************************************/

void PetriNet::generateCachePartition(const ModeConfiguration::Config& config, const int partition_id,
                                      const int partitions_count, const bool screen_output) {
    tracker.reset();
    EnzymesTokensEncoder *encoder = new BinaryEnzymesTokensEncoder(enzymes);
    ofstream cacheFile;
    std::ostringstream cachePath;
    cachePath << "/tmp/cache_" << partition_id << "_" << partitions_count << ".txt";
    cacheFile.open(cachePath.str());
    int max = encoder->getMaximalEncoding();
    int encodings_per_partition = (max / partitions_count) + 1;
    int start = partition_id * encodings_per_partition;
    int end = (partition_id + 1) * encodings_per_partition;
    for (int i = start; i < end && i < encoder->getMaximalEncoding(); ++i) {
        if (!screen_output) cout << "Caching " << i << " from range " << start << " - " << end << "\n";
        encoder->decodeAndSetSubstancesStates(i);
        runMetabolismStep(config);
        cacheFile << i << " ";
        if (screen_output) cout << i << " ";
        for (int metaboliteInd = 0; metaboliteInd < Nm; ++metaboliteInd) {
            double state = metabolites[metaboliteInd]->getState();
            std::string name = metabolites[metaboliteInd]->getName();
            cacheFile << name << " " << state << " ";
            if (screen_output) cout << name << " " << state << " ";
        }
        cacheFile << "\n";
        if (screen_output) cout << "\n";
    }
    cacheFile.close();
}

string PetriNet::runSingleStep(const ModeConfiguration::Config& config) {
    tracker.reset();
	loadMetabolismStep(config);

    updateResetReactions();
    updateInstantReactions();
    string rname = hybridUpdate3();
    if (config.mode == ModeConfiguration::STANDALONE) {
        gsmnDriver->setUpdateRequired(false);
    }

    int fgf_track_id = substances[nameIndex[COMMUNICATOR]].get_tracking_id();
    int prod = tracker.produced(fgf_track_id);
    int cons = tracker.consumed(fgf_track_id);
    if (prod != 0 || cons != 0) {
      cout << COMMUNICATOR << " tracking: +" << prod << ", -" << cons << endl;
    }
    return rname;

}

void PetriNet::preTrajectorySetup() {
  Time = 0;
  deltaTime = 0;
  timeSinceNutrientUpdate = 0;
  deltaTime = maxTimestep;
  trajectories.push_back(Trajectory());
  runStateMonitors(trajectories.size()-1, "none", Time);
}

void PetriNet::clearTrajectories() {
  trajectories.clear();
}
bool PetriNet::runSingleTrajectory(const ModeConfiguration::Config& config) {
    preTrajectorySetup();
    int trajectoryIdx = trajectories.size()-1;

    for(int iteration_counter = 1; Time < TimeMax; iteration_counter++) {
        if (target_flag && substances[targetIdx].getState() >= targetThreshold) {
            return true;
        }
        string rname = runSingleStep(config);
        if (iteration_counter % monitor_fq == 0) {
          runStateMonitors(trajectoryIdx, rname, Time);
        }
        if (progress_flag && iteration_counter % progress_fq == 0) cout << "Trajectory: " << trajectoryIdx << " Time: " << Time << endl;
    }
    return false;
}

void PetriNet::oneTimeSetup(const string& ctrl_file, const ModeConfiguration::Config& config) {
  readControlFile(ctrl_file);
  /* Make sfba path relative to control file: */
  size_t last_slash = ctrl_file.find_last_of("/");
  if (last_slash != string::npos) {
    sfba_file = ctrl_file.substr(0, last_slash) + "/" + sfba_file;
  }

  buildGsmnDriver();
  if (config.mode == ModeConfiguration::SERVER) {
    return;
  }


  if (config.mode == ModeConfiguration::CLIENT) {
    gsmnDriver->set_server_address(config.serverAddress, config.serverPort);
  }
  prepareForSampling(Nsample, Seed);
  splitReactions();
  initialState = getState();
  /* Currently we track only one communicator, and depend on it having idx 0. */
  int track_id = substances[nameIndex[COMMUNICATOR]].reportTo(&tracker);
  cout << "Tracking " << COMMUNICATOR << " (" << nameIndex[COMMUNICATOR] << ") with track idx " << track_id << endl;
}

void PetriNet::hybridSimulation3(const ModeConfiguration::Config& config) {
    int i, Ntarget;
    string rname;
    ofstream progfile;

    /** TEST: */

    Ntarget = 0;
    for (i = 0; i < Nsample; i++) {
        setState(initialState);
        bool reachesTarget = runSingleTrajectory(config);
        if (reachesTarget) {
          Ntarget++;
        }

        progfile.open(logfile, ofstream::app);
        progfile << "TRAJECTORY " << i << endl;
        progfile.close();


    }

    progfile.open(logfile.c_str(), ofstream::app);
    progfile << "NUMBER OF TRAJECTORIES REACHING TARGET: " << Ntarget << endl;
    progfile << "TOTAL NUMBER OF TRAJECTORIES: " << Nsample << endl;
    progfile << "RATIO OF TRAJECTORIES REACHING TARGET: " << (float) Ntarget / (float) Nsample << endl;
    progfile << endl;
    progfile << VERSION;
    progfile.close();

    for (i = 0; i < Nsample; i++) {
        trajectories[i].printEventAmountTableWithTime(outfile, i);
    }
}

double PetriNet::runOptimization(std::string clientString) {
    if (gsmnDriver == NULL) {
      buildGsmnDriver();
    } else {
      gsmnDriver->reset_changes();
    }
    gsmnDriver->restoreChangesMapAndUpdateLP(clientString);
    return gsmnDriver->calcObjValue();
}



void PetriNet::tokenGame1() {
    int i, Ntarget;
    string rname;
    ofstream progfile;

    Ntarget = 0;
    if (msg_flag) cout << "ALLOCATE TRAJECTORIES AND INITIALISE RNG" << endl;
    prepareForSampling(Nsample, Seed);
    if (msg_flag) cout << "SPLIT FAST AND SLOW REACTIONS" << endl;
    splitReactions();

    if (msg_flag) cout << "RECORD INITIAL STATE OF PETRI NET" << endl;
    initialState = getState();

    trajectories.clear();
    trajectories.resize(Nsample);

    for (i = 0; i < Nsample; i++) {
        Time = 0;
        deltaTime = 0;

        if (msg_flag) cout << endl << "*********** TRAJECTORY " << i << " ***********" << endl;
        if (msg_flag) cout << "Reload initial state of Petri Net" << endl;
        setState(initialState);

        deltaTime = maxTimestep;

        while (Time < TimeMax) {
            if (msg_flag) cout << endl << "ITERATION " << Time << endl;
            if (substances[targetIdx].getState() >= targetThreshold) {
                Ntarget++;
                break;
            }

            if (msg_flag) cout << "### Update fast reactions" << endl;
            updateFastReactions1();
            if (msg_flag) cout << "### Fire discrete event:" << endl;
            rname = hybridUpdate1();

            if (msg_flag) cout << Time << "\t" << rname << endl;
            if (msg_flag) cout << "RECORD THE STATE AT TIME " << Time << endl;
            runStateMonitors(i, rname, Time);
        }

        if (i % monitor_fq == 0) {
            progfile.open(logfile.c_str(), ofstream::app);
            progfile << "TRAJECTORY " << i << endl;
            progfile.close();
        }
    }

    progfile.open(logfile.c_str(), ofstream::app);
    progfile << "NUMBER OF TRAJECTORIES REACHING TARGET: " << Ntarget << endl;
    progfile << "TOTAL NUMBER OF TRAJECTORIES: " << Nsample << endl;
    progfile << "RATIO OF TRAJECTORIES REACHING TARGET: " << (float) Ntarget / (float) Nsample << endl;
    progfile << endl;
    progfile << VERSION;
    progfile.close();

    for (i = 0; i < Nsample; i++) {
        trajectories[i].printEventAmountTableWithTime(outfile, i);
    }

}

/************************************************************************************/

void RandomNumberGenerator::rmarzam(int newseed) {

    /* initialization variables */
    int ij, kl, i, j, k, l, m, ii, jj;
    float s, t;

    /* random number variables */
    int seed; /* default seed if none specified */

    seed = newseed % 900000000;
    ij = seed / 30082;
    kl = seed - 30082 * ij;
    i = ((ij / 177) % 177) + 2;
    j = (ij % 177) + 2;
    k = ((kl / 169) % 178) + 1;
    l = (kl % 169);

    for (ii = 0; ii < 97; ii++) {

        s = 0.0;
        t = 0.5;

        for (jj = 0; jj < 24; jj++) {

            m = (((i * j) % 179) * k) % 179;
            i = j;
            j = k;
            k = m;
            l = ((53 * l) + 1) % 169;
            if (((l * m) % 64) >= 32)
                s += t;
            t *= .5;
        }

        u[ii] = s;
    }

    c = 362436. / 16777216.;
    cd = 7654321. / 16777216.;
    cm = 16777213. / 16777216.;
    i96 = 96;
    j96 = 32;
}

double RandomNumberGenerator::marzam() {
    /* random number generation */
    uni = u[i96] - u[j96];
    if (uni < 0.) uni += 1.0;
    u[i96] = uni;
    i96--;
    if (i96 < 0) i96 = 96;
    j96--;
    if (j96 < 0) j96 = 96;
    c -= cd;
    if (c < 0.) c += cm;
    uni -= c;
    if (uni < 0.) uni += 1.0;

    return ( (double) uni);
}
