package biosim

import biosim.Simulation._
import biosim.qsspn.ThriftConnection
import org.apache.spark.graphx.{EdgeContext, Graph, TripletFields}
import org.apache.spark.util.LongAccumulator
import thrift.ModelDescription
import org.log4s._

import scala.collection.immutable.HashMap
import scala.util.Random

trait Simulation {
  def modelDescription: ModelDescription
  def bcastFieldIdx: Int
  def pRunSingleCell(cellState: PCellSt, usedTokens: LongAccumulator): PCellSt

  def stepSystem(world: Simulation.PrecWorld, usedTokens: LongAccumulator): Simulation.PrecWorld
}

case class PCellSt(
                    id: Int,
                    curState: Array[Boolean],
                    nextState: Array[Boolean],
                    fstStep: Boolean,
                    canGive: Boolean,
                    offers: List[Int],
                    requests: Requests,
                    request: Option[Int],
                    offer: Option[Int]
)

object Simulation {
  type BinaryMetadata = HashMap[String, Boolean]
  type CellState = (Array[Boolean], BinaryMetadata)
  type World = Graph[CellState, Nothing]
  type PrecWorld = Graph[PCellSt, Nothing]
  type Requests = List[(Int, Int)]
}

class SimulationImpl(qsspn: ThriftConnection, bcastField: String, opts: HashMap[String, Boolean]) extends Simulation with Serializable {
  lazy val modelDescription = qsspn.modelDescription
  lazy val bcastFieldIdx = modelDescription.placeNames.zipWithIndex.toMap.apply(bcastField)

  private[this] val log = getLogger

  lazy private[this] val sourceOnly = new TripletFields(true, false, false)


  def stepSystem(world: Simulation.PrecWorld, usedTokens: LongAccumulator): Simulation.PrecWorld = {
    System.gc()
    val tokens = world.vertices.filter { case (id, state) => state.curState(bcastFieldIdx) }.count()
    log info "System step - " + tokens + " tokens available, " + usedTokens + " tokens used"
    val oldStep = world.vertices.filter { case (id, state) => state.fstStep }.isEmpty()
    val allServed = world.vertices.filter { case (id, state) => state.request.isDefined }.isEmpty()

    if (oldStep && allServed) moveAllCellsToNextState(world, usedTokens)
    else runCommunicationStep(world, oldStep, usedTokens)
  }

  def moveAllCellsToNextState(world: PrecWorld, usedTokens: LongAccumulator): Graph[PCellSt, Nothing] = {
    log info "Transition step - " + usedTokens + " tokens used"
    usedTokens.reset()
    world.mapVertices {
      case (id, state) => state.copy(curState = state.nextState, fstStep = true,
        offers = List(), requests = List(), request = None, offer = None)
        //TODO canGive = false
    }
  }


  def runCommunicationStep(world: PrecWorld, oldStep: Boolean, usedTokens: LongAccumulator): PrecWorld = {
    log info "Communication step"
    val offers = world.aggregateMessages[List[Int]](trio => trio.sendToDst(offId(trio.srcAttr)), _ ::: _, sourceOnly)
    val requests = world.aggregateMessages[Requests](trio => trio.sendToDst(reqId(trio.srcAttr)), _ ::: _, sourceOnly)
    val donations = world.aggregateMessages[Boolean](trio => trio.sendToDst(give(trio)), _ || _)

    val withOffs = world.outerJoinVertices(offers) { (id, s, offers) => addOffs(s, offers) }
    val withOffsReqs = withOffs.outerJoinVertices(requests) { (id, s, requests) => addReqs(s, requests) }
    val completeWorld = withOffsReqs.outerJoinVertices(donations) { (id, s, d) => addDonation(s, d.getOrElse(false)) }

    val nextWorld = completeWorld.mapVertices { case (id, state) => pRunSingleCell(state, usedTokens) }

    world.unpersist()
    withOffs.unpersist()
    withOffsReqs.unpersist()
    completeWorld.unpersist()
    stepSystem(nextWorld, usedTokens)
  }

  def addOffs(cs: PCellSt, offs: Option[List[Int]]): PCellSt = cs.copy(offers = offs.getOrElse(List[Int]()))
  def addReqs(cs: PCellSt, reqs: Option[Requests]): PCellSt = cs.copy(requests = reqs.getOrElse(List[(Int, Int)]()))
  def addDonation(cs: PCellSt, donation: Boolean): PCellSt = cs.copy(offer = if (donation) Some(0) else None)

  def offId(c: PCellSt): List[Int] = if (canGive(c)) List(c.id) else List[Int]()
  def reqId(c: PCellSt): Requests = if (c.request.isDefined) List((c.id, c.request.get)) else List()
  def give(trio: EdgeContext[PCellSt, Nothing, Boolean]): Boolean = trio.srcAttr.offer.contains(trio.dstAttr.id)

  def canGive(c: PCellSt): Boolean = (c.curState(bcastFieldIdx) && c.fstStep) || (!c.fstStep && c.canGive)

  def pRunSingleCell(cs: PCellSt, usedTokens: LongAccumulator): PCellSt =
    if (cs.fstStep) pRunFstAttempt(cs, usedTokens)
    else pRunNextAttempt(cs, usedTokens)

  def pRunFstAttempt(cs: PCellSt, usedTokens: LongAccumulator): PCellSt = {
    val hadOwn = cs.curState(bcastFieldIdx)
    cs.curState(bcastFieldIdx) |= cs.offers.nonEmpty

    val stepResult = qsspn.requestStep(cs.curState)
    val resultMarking = stepResult.marked
    val produced = (modelDescription.trackedPlaceIndexes, stepResult.produced).zipped.toMap

    val used = (hadOwn || cs.offers.nonEmpty) && (!resultMarking(bcastFieldIdx) || produced(bcastFieldIdx))
    if (used) usedTokens.add(1)

    cs.copy(nextState = resultMarking,
      fstStep = false,
      canGive = hadOwn && !used,
      request = if (used && !hadOwn) Option(Random.shuffle(cs.offers).head) else None)
  }

  def pRunNextAttempt(cs: PCellSt, usedTokens: LongAccumulator): PCellSt = {
    val csWithOffer = if (cs.canGive) {
      val askedMe = cs.requests.dropWhile( msg => msg._2 != cs.id ).map( msg => msg._1)

      if (askedMe.isEmpty) cs.copy(offer = None)
      else {
        cs.nextState(bcastFieldIdx) = false
        cs.copy(canGive = false, offer = Option(Random.shuffle(askedMe).head))
      }
    } else cs.copy(offer = None)

    val askedAndDidNotGet = cs.offer.isEmpty && cs.request.isDefined

    if (askedAndDidNotGet) {
      if (cs.offers.isEmpty) pStepWithoutCommunicator(csWithOffer)
      else csWithOffer.copy(request = Option(Random.shuffle(cs.offers).head))
    } else { // did not ask or did get
      if (cs.request.isDefined) {
        usedTokens.add(1)
        csWithOffer.copy(request = None)
      } // get
      else csWithOffer // did not ask
    }
  }

  def pStepWithoutCommunicator(cs: PCellSt): PCellSt = {
    cs.curState(bcastFieldIdx) = false

    val stepResult = qsspn.requestStep(cs.curState)
    val resultMarking = stepResult.marked

    cs.copy(nextState = resultMarking, request = None, offer = None)
  }

}