package biosim.graphs

import scala.reflect._
import org.apache.spark._
import org.apache.spark.graphx._

import scala.collection.immutable.HashMap

/** Creates graphs for simulations.
 *  Provides factory methods capable of creating
 *  instances of Driver.World class, along
 *  with biosim.core.setting.Grid instances for description.
 */
object GraphBuilder {
  type DescribedWorld[CellState] = (Graph[CellState, Nothing], Grid)

  def isolated[CS](size: Int, cs: CS)(implicit sc: SparkContext, cstag: ClassTag[CS]): DescribedWorld[CS] = {
    val vertices = sc.parallelize(for {id <- 1L to size} yield (id, cs))
    val edges: rdd.RDD[Edge[Nothing]] = sc.parallelize(Seq())
    val graph: Graph[CS, Nothing] = Graph[CS, Nothing](vertices, edges, cs)
    (graph, Isolated(size))
  }
  def lined[CS](size: Int, csMaker: (Long => CS))(implicit sc: SparkContext, cstag: ClassTag[CS]): DescribedWorld[CS] = {
    val vertices = sc.parallelize(for {id <- 1L to size} yield(id, csMaker(id)))
    val edges = for {
      id <- 1L to (size-1)
    } yield Seq(Edge(id, id+1), Edge(id+1, id))
    val graph: Graph[CS, Nothing] = Graph[CS, Nothing](vertices, sc.parallelize(edges.flatten), csMaker(0))
    (graph, Lined(size))
  }

  def loopCube[A](from: Long, to: Long)(generator: (Long, Long, Long) => A) = for {
    x <- from to to
    y <- from to to
    z <- from to to
  } yield generator(x, y, z)

  val regularNeighbourhood = (x: Long, y: Long, z: Long) => {
    Seq(
      ((x, y, z), (x, y, z+1)),
      ((x, y, z), (x, y+1, z)),
      ((x, y, z), (x+1, y, z)))
  }

  val extendedNeighbourhood = (x: Long, y: Long, z: Long) => {
    Seq(
      ((x, y, z), (x, y, z + 1)),
      ((x, y, z), (x, y + 1, z)),
      ((x, y, z), (x + 1, y, z)),
      ((x, y, z), (x, y + 1, z + 1)),
      ((x, y, z), (x + 1, y, z + 1)),
      ((x, y, z), (x + 1, y + 1, z)),
      ((x, y, z), (x + 1, y + 1, z + 1)))
  }

  def cube[CS](size: Int, csMaker: (Long, Long, Long) => CS, opts: HashMap[String, Boolean])
              (implicit sc: SparkContext, cstag: ClassTag[CS]): DescribedWorld[CS] = {
    val base = size+1
    def toIdx(coords: (Long, Long, Long)) = base*base*coords._1 + base*coords._2 + coords._3

    val vertices = loopCube(1, size) {(x, y, z) => (toIdx(x,y,z), csMaker(x, y, z))}
    val neighbourhood = if(opts("extendedNeighbourhood"))
      extendedNeighbourhood
    else
      regularNeighbourhood

    val edges = loopCube(1, size) { (x,y,z) => neighbourhood(x, y, z)
      .filter { case (_, (a,b,c)) => (a max b max c) <= size }
      .flatMap { case (a, b) => Seq((a, b), (b, a)) }
      .map {case (a, b) => Edge(toIdx(a), toIdx(b))}
    }

    val graph: Graph[CS, Nothing] = Graph[CS, Nothing](sc.parallelize(vertices), sc.parallelize(edges.flatten), csMaker(0, 0, 0))
    (graph, Cube(size))
  }

  case class Isolated(override val size: Int) extends Grid
  case class Lined(override val size: Int) extends Grid
  case class Cube(singleDimension: Int) extends Grid {
    override val size = singleDimension * singleDimension * singleDimension
    override def toString = s"Cube ($singleDimension x $singleDimension x $singleDimension = $size cells)"
  }

}
