package biosim.spark

import biosim.{PCellSt, Simulation}
import biosim.Simulation.PrecWorld
import org.apache.spark.SparkContext
import org.apache.spark.graphx.{Graph, PartitionStrategy, VertexId}
import org.apache.spark.util.LongAccumulator
import org.log4s._


class Runner(simulation: Simulation) {

  val log = getLogger

  val homeostasisIdx = simulation.modelDescription.placeNames.indexOf("BA_HOMEOSTASIS")
  val homeostasisTracker = new HomeostasisTracker(homeostasisIdx)

  def countPartitions[A, B](world: Graph[A, B]) = (world.vertices.partitions.length, world.edges.partitions.length)

  def run(world: PrecWorld, steps: Int, usedTokens: LongAccumulator)(implicit sc: SparkContext): PrecWorld = {
    var currentWorld = world

    currentWorld = currentWorld.partitionBy(PartitionStrategy.EdgePartition2D)
    log info s"Starting simulation with $steps steps"

    var nextWorld = currentWorld
    var homeostasisLog = sc.parallelize(Seq[(VertexId, Int)]())

    1 to steps foreach { stepNo =>
      nextWorld = simulation.stepSystem(currentWorld, usedTokens)
      currentWorld.unpersist()
      if (stepNo+1 == steps) {
        nextWorld = nextWorld
      }

      val newHomeostasis = currentWorld.vertices
        .join(nextWorld.vertices)
        .filter(homeostasisTracker.newOccurrence)
        .map(homeostasisTracker.tagStepNumber(stepNo))

      homeostasisLog = homeostasisLog.union(newHomeostasis)
      currentWorld.unpersist()
      currentWorld = nextWorld

    }
    log info "Simulation stage constructed, starting up"
    log info homeostasisLog.collect().mkString("Homeostasis log: \n", "\n", "\n")
    currentWorld
  }
}

/**
  * Helper classes that encapsulate spark RDD operations. When using methods as spark RDD transformation arguments,
  * the whole enclosing class is serialized - we want to avoid serializing the whole Runner.
  */

class FieldExtractor(idx: Int) extends (PCellSt => Boolean) with Serializable {
  override def apply(state: PCellSt): Boolean = state.curState(idx)
}

class HomeostasisTracker(idx: Int) extends Serializable {
  val extractHomeostasis = new FieldExtractor(idx)
  val newOccurrence: ((VertexId, (PCellSt, PCellSt))) => Boolean = {
    case (_, (oldState, newState)) => !extractHomeostasis(oldState) && extractHomeostasis(newState)
  }
  def tagStepNumber(stepNo: Int): ((VertexId, Any)) => (VertexId, Int) = {
    case (id, _) => (id, stepNo)
  }
}
