package biosim

import biosim.Simulation.{CellState, World}
import biosim.checkpoints.CheckpointSaver
import biosim.graphs.{GraphBuilder, Grid}
import biosim.qsspn.ThriftConnection
import biosim.spark.Runner
import org.apache.spark.graphx.Graph
import org.apache.spark.{SparkConf, SparkContext}
import org.log4s._
import com.twitter.conversions.time._

import scala.collection.immutable.HashMap
import scala.util.Try


object Main {

  val COMMUNICATOR = "FGF19"

  def main(args: Array[String]) {
    val optionalParameters = "[communicatorsLiveOneTurn] [turnOffUsedCommunicators] [extraCommunicationPass] [extendedNeighbourhood] [noCommunication]"
    require(args.length >= 3, "Usage: Runner size maxSteps pathToCheckpointDir " + optionalParameters)

    implicit val sc = sparkInit
    val size = args(0).toInt
    val maxSteps = args(1).toInt
    val checkpointSaver = new CheckpointSaver(args(2))

    val opts: HashMap[String, Boolean] = collection.immutable.HashMap(
      "communicatorsLiveOneTurn" -> Try(args(3).toBoolean).getOrElse(false),
      "turnOffUsedCommunicators" -> Try(args(4).toBoolean).getOrElse(false),
      "extraCommunicationPass" -> Try(args(5).toBoolean).getOrElse(false),
      "extendedNeighbourhood" -> Try(args(6).toBoolean).getOrElse(false),
      "noCommunication" -> Try(args(7).toBoolean).getOrElse(false)
    )

    val (world, grid) = worldInit(size, opts)
    log info s"Constructed grid: $grid"
    var currentWorld = world
    var currentStep = 0
    checkpointSaver.restoreVertices().foreach {
      case (vertices, step) => {
        currentWorld = Graph[CellState, Nothing](vertices, currentWorld.edges)
        currentStep = step
        log info s"Loaded graph with ${vertices.count()} vertices from step $step"
      }
    }

    val runner = runnerInit(opts)

    while (currentStep < maxSteps) {
      currentWorld = runner.run(currentWorld, stageSize)
      currentStep += stageSize
      checkpointSaver.save(currentWorld, currentStep)
      log info s"Saved world checkpoint after step $currentStep"
    }
    log info "All done, shutting down"
  }

  val log = getLogger
  val stageSize = 10

  def sparkInit = {
    val conf = new SparkConf()
      .setAppName("Homeostasis simulation")
      .set("spark.eventLog.enabled", "true")
    new SparkContext(conf)
  }

  def worldInit(size: Int, opts: HashMap[String, Boolean])(implicit sc: SparkContext): (World, Grid)= {
    val initialState = (List(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0,
      0, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0,
      0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1,
      0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0).map(_ == 1).toArray,
      HashMap("consumedOuterCommunicator" -> false, "hasAccessToCommunicator" -> false))

    val world = GraphBuilder.cube(size, (x, y, z) => initialState, opts)
    log info s"Constructed grid: ${world._2}"
    world
  }

  def runnerInit(opts: HashMap[String, Boolean]): Runner = {
    val qsspn = ThriftConnection("127.0.0.1", 9090, 2.minutes)
    val simulation: Simulation = new SimulationImpl(qsspn, COMMUNICATOR, opts)
    new Runner(simulation)
  }



}

