package biosim

import biosim.Simulation.{CellState, World}
import biosim.qsspn.ThriftConnection
import org.apache.spark.graphx.{Graph, TripletFields}
import thrift.ModelDescription
import org.log4s._

import scala.collection.immutable.HashMap

trait Simulation {
  def modelDescription: ModelDescription
  def bcastFieldIdx: Int
  def runSingleCell(cellState: CellState): CellState

  def stepSystem(world: Simulation.World): Simulation.World
}

object Simulation {
  type CellState = (Array[Boolean], HashMap[String, Boolean])
  type World = Graph[CellState, Nothing]
}

class SimulationImpl(qsspn: ThriftConnection, bcastField: String, opts: HashMap[String, Boolean]) extends Simulation with Serializable{
  lazy val modelDescription = qsspn.modelDescription
  lazy val bcastFieldIdx = modelDescription.placeNames.zipWithIndex.toMap.apply(bcastField)

  private[this] val log = getLogger

  def runSingleCell(cellState: CellState): CellState = {
    val state = cellState._1
    val metadata = cellState._2

    if (opts("noCommunication")) {
      val stepResult = qsspn.requestStep(state)
      val resultMarking = stepResult.marked
      (resultMarking, metadata)
    } else {
      val env = metadata("hasAccessToCommunicator")
      val ownBcastValue = state(bcastFieldIdx) && (!metadata("sacrificedCommunicator") || !opts("turnOffUsedCommunicators"))
      state(bcastFieldIdx) |= env
      val stepResult = qsspn.requestStep(state)
      val resultMarking = stepResult.marked
      val produced = (modelDescription.trackedPlaceIndexes, stepResult.produced).zipped.toMap
      val consumedOuterCommunicator = !ownBcastValue && env && !stepResult.marked(bcastFieldIdx)
      resultMarking(bcastFieldIdx) = (!opts("communicatorsLiveOneTurn") && ownBcastValue) || produced(bcastFieldIdx)
      (resultMarking, metadata + ("consumedOuterCommunicator" -> consumedOuterCommunicator))
    }
  }

  lazy private[this] val sourceOnly = new TripletFields(true, false, false)
  def stepSystem(world: Simulation.World): Simulation.World = {
    if (opts("noCommunication")) {
      world.mapVertices {
        case (id, state) => runSingleCell(state)
      }
    } else {
      val sacrifices = world.aggregateMessages[Boolean](
        triplet => triplet.sendToDst(triplet.srcAttr._2("consumedOuterCommunicator")),
        _ || _,
        sourceOnly
      )
      val worldWithSacrifices: World = world.outerJoinVertices(sacrifices) { (id, state, sacrifice) => (state._1, state._2 + ("sacrificedCommunicator" -> sacrifice.getOrElse(false))) }
      val messages = world.aggregateMessages[Boolean](
        triplet => triplet.sendToDst(triplet.srcAttr._1(bcastFieldIdx) && !triplet.srcAttr._2("sacrificedCommunicator")),
        _ || _,
        sourceOnly
      )
      val envedWorld: World = worldWithSacrifices.outerJoinVertices(messages) { (id, state, env) => (state._1, state._2 + ("hasAccessToCommunicator" -> env.getOrElse(false)))}
      val actualWorld = if (opts("extraCommunicationPass")) {
        val messages = envedWorld.aggregateMessages[Boolean](
          triplet => triplet.sendToDst(triplet.srcAttr._2("hasAccessToCommunicator") && !triplet.srcAttr._2("sacrificedCommunicator")),
          _ || _,
          sourceOnly
        )
        envedWorld.outerJoinVertices(messages) { (id, state, env) => (state._1, state._2 + ("hasAccessToCommunicator" -> env.getOrElse(false)))}
      } else {
        envedWorld
      }
      actualWorld.mapVertices {
        case (id, state) => runSingleCell(state)
      }
    }
  }
}




