package biosim.qsspn

import thrift._

import com.twitter.finagle.Thrift
import com.twitter.util.{Await, Duration}

//import scala.concurrent.Await
//import scala.concurrent.duration.Duration

import scala.collection.mutable.{HashMap, Map}

trait ThriftConnection {

  val modelDescription: ModelDescription

  /**
   * Requests a single step of computation from underlying QSSPN server.
   * @param cellState current marking.
   * @return new marking. Second value is true if the cell produced its own token of broadcast field.
   */
  def requestStep(cellState: Seq[Boolean]): StepResult

  /**
   * Represents a result of single simulation step.
   * @param marked marking of the Petri net.
   * @param produced production log of some of the places. To check which places are
   *                 being tracked, use `describeModel`.
   */
  case class StepResult(marked: Array[Boolean], produced: Seq[Boolean])

}

/**
 * A factory object for Thrift connections.
 * It opens only one connection for every specified thrift server.
 */
object ThriftConnection {

  private[this] val connectionMap: Map[(String, Int), TIface] = HashMap.empty

  def apply(srv_addr: String, srv_port: Int, timeout: Duration): ThriftConnection = {
    new LazyThriftConnection(srv_addr, srv_port, timeout)
  }

  private def getIface(srv_addr: String, srv_port: Int): TIface = {
    connectionMap.getOrElseUpdate((srv_addr, srv_port), {
      val target = s"${srv_addr}:${srv_port}"
      Thrift.newIface[TIface](target)
    })
  }
  type TIface = SingleStepProxy.FutureIface

  private class LazyThriftConnection(val srv_addr: String, val srv_port: Int, val timeout: Duration) extends ThriftConnection with Serializable {
    override lazy val modelDescription: ModelDescription = Await.result(client.describeModel(), timeout)

    override def requestStep(cellState: Seq[Boolean]): StepResult = {
      val marking = Marking(cellState.toList)
      val result = client.calculateStep(marking)

      val res_marking = Await.result(result, timeout)
      StepResult(res_marking.marked.toArray, res_marking.produced)
    }

    def client = ThriftConnection.getIface(srv_addr, srv_port)
  }

}
